@@ -19,6 +19,7 @@ from __future__ import print_function
 
 __metaclass__ = type
 __all__ = [
+    'ClickBuildError',
     'ClickBuilder',
     'ClickSourceBuilder',
     ]
@@ -32,6 +33,7 @@ import os
 import re
 import shutil
 import subprocess
+import sys
 import tarfile
 import tempfile
 from textwrap import dedent
@@ -62,6 +64,10 @@ class FakerootTarFile(tarfile.TarFile):
         return tarinfo
 
 
+class ClickBuildError(Exception):
+    pass
+
+
 class ClickBuilderBase:
     def __init__(self):
         self.file_map = {}
@@ -71,7 +77,18 @@ class ClickBuilderBase:
 
     def read_manifest(self, manifest_path):
         with io.open(manifest_path, encoding="UTF-8") as manifest:
-            self.manifest = json.load(manifest)
+            try:
+                self.manifest = json.load(manifest)
+            except Exception as e:
+                raise ClickBuildError(
+                    "Error reading manifest from %s: %s" % (manifest_path, e))
+            keys = sorted(self.manifest)
+            for key in keys:
+                if key.startswith("_"):
+                    print(
+                        "Ignoring reserved dynamic key '%s'." % key,
+                        file=sys.stderr)
+                    del self.manifest[key]
 
     @property
     def name(self):
@@ -95,10 +112,47 @@ class ClickBuilderBase:
 
     @property
     def architecture(self):
-        return self.manifest.get("architecture", "all")
+        manifest_arch = self.manifest.get("architecture", "all")
+        if isinstance(manifest_arch, list):
+            return "multi"
+        else:
+            return manifest_arch
 
 
 class ClickBuilder(ClickBuilderBase):
+    # TODO: This should be configurable, or at least extensible.
+    _ignore_patterns = [
+        ".*.sw?",
+        "*~",
+        ",,*",
+        ".[#~]*",
+        ".arch-ids",
+        ".arch-inventory",
+        ".be",
+        ".bzr",
+        ".bzr-builddeb",
+        ".bzr.backup",
+        ".bzr.tags",
+        ".bzrignore",
+        ".cvsignore",
+        ".git",
+        ".gitattributes",
+        ".gitignore",
+        ".gitmodules",
+        ".hg",
+        ".hgignore",
+        ".hgsigs",
+        ".hgtags",
+        ".shelf",
+        ".svn",
+        "CVS",
+        "DEADJOE",
+        "RCS",
+        "_MTN",
+        "_darcs",
+        "{arch}",
+        ]
+
     def list_files(self, root_path):
         for dirpath, _, filenames in os.walk(root_path):
             rel_dirpath = os.path.relpath(dirpath, root_path)
@@ -113,6 +167,27 @@ class ClickBuilder(ClickBuilderBase):
             return None
         return tarinfo
 
+    def _pack(self, temp_dir, control_dir, data_dir, package_path):
+        data_tar_path = os.path.join(temp_dir, "data.tar.gz")
+        with contextlib.closing(FakerootTarFile.open(
+                name=data_tar_path, mode="w:gz", format=tarfile.GNU_FORMAT
+                )) as data_tar:
+            data_tar.add(data_dir, arcname="./", filter=self._filter_dot_click)
+
+        control_tar_path = os.path.join(temp_dir, "control.tar.gz")
+        control_tar = tarfile.open(
+            name=control_tar_path, mode="w:gz", format=tarfile.GNU_FORMAT)
+        control_tar.add(control_dir, arcname="./")
+        control_tar.close()
+
+        with ArFile(name=package_path, mode="w") as package:
+            package.add_magic()
+            package.add_data("debian-binary", b"2.0\n")
+            package.add_data(
+                "_click-binary", ("%s\n" % spec_version).encode("UTF-8"))
+            package.add_file("control.tar.gz", control_tar_path)
+            package.add_file("data.tar.gz", data_tar_path)
+
     def build(self, dest_dir, manifest_path="manifest.json"):
         with make_temp_dir() as temp_dir:
             # Prepare data area.
@@ -122,22 +197,19 @@ class ClickBuilder(ClickBuilderBase):
                 if dest_path.startswith("/"):
                     dest_path = dest_path[1:]
                 real_dest_path = os.path.join(root_path, dest_path)
-                shutil.copytree(source_path, real_dest_path, symlinks=True)
+                shutil.copytree(
+                    source_path, real_dest_path, symlinks=True,
+                    ignore=shutil.ignore_patterns(*self._ignore_patterns))
 
             # Prepare control area.
             control_dir = os.path.join(temp_dir, "DEBIAN")
             osextras.ensuredir(control_dir)
 
-            # Control file names must not contain a dot, hence "manifest"
-            # rather than "manifest.json" in the control area.
-            real_manifest_path = os.path.join(control_dir, "manifest")
             if os.path.isabs(manifest_path):
                 full_manifest_path = manifest_path
             else:
                 full_manifest_path = os.path.join(root_path, manifest_path)
-            os.rename(full_manifest_path, real_manifest_path)
-            os.chmod(real_manifest_path, 0o644)
-            self.read_manifest(real_manifest_path)
+            self.read_manifest(full_manifest_path)
 
             du_output = subprocess.check_output(
                 ["du", "-k", "-s", "--apparent-size", "."],
@@ -146,6 +218,7 @@ class ClickBuilder(ClickBuilderBase):
             if not match:
                 raise Exception("du gave unexpected output '%s'" % du_output)
             installed_size = match.group(1)
+            self.manifest["installed-size"] = installed_size
             control_path = os.path.join(control_dir, "control")
             osextras.ensuredir(os.path.dirname(control_path))
             with io.open(control_path, "w", encoding="UTF-8") as control:
@@ -161,6 +234,19 @@ class ClickBuilder(ClickBuilderBase):
                     self.maintainer, installed_size, self.title)),
                     file=control)
 
+            # Control file names must not contain a dot, hence "manifest"
+            # rather than "manifest.json" in the control area.
+            real_manifest_path = os.path.join(control_dir, "manifest")
+            with io.open(
+                    real_manifest_path, "w", encoding="UTF-8") as manifest:
+                print(
+                    json.dumps(
+                        self.manifest, ensure_ascii=False, sort_keys=True,
+                        indent=4, separators=(",", ": ")),
+                    file=manifest)
+            os.unlink(full_manifest_path)
+            os.chmod(real_manifest_path, 0o644)
+
             md5sums_path = os.path.join(control_dir, "md5sums")
             with open(md5sums_path, "w") as md5sums:
                 for path in sorted(self.list_files(root_path)):
@@ -178,27 +264,10 @@ class ClickBuilder(ClickBuilderBase):
                 preinst.write(static_preinst)
 
             # Pack everything up.
-            data_tar_path = os.path.join(temp_dir, "data.tar.gz")
-            with contextlib.closing(FakerootTarFile.open(
-                    name=data_tar_path, mode="w:gz", format=tarfile.GNU_FORMAT
-                    )) as data_tar:
-                data_tar.add(
-                    root_path, arcname="./", filter=self._filter_dot_click)
-
-            control_tar_path = os.path.join(temp_dir, "control.tar.gz")
-            control_tar = tarfile.open(
-                name=control_tar_path, mode="w:gz", format=tarfile.GNU_FORMAT)
-            control_tar.add(control_dir, arcname="./")
-            control_tar.close()
-
             package_name = "%s_%s_%s.click" % (
                 self.name, self.epochless_version, self.architecture)
             package_path = os.path.join(dest_dir, package_name)
-            with ArFile(name=package_path, mode="w") as package:
-                package.add_magic()
-                package.add_data("debian-binary", b"2.0\n")
-                package.add_file("control.tar.gz", control_tar_path)
-                package.add_file("data.tar.gz", data_tar_path)
+            self._pack(temp_dir, control_dir, root_path, package_path)
             return package_path
 
 
