@@ -18,8 +18,10 @@
 from __future__ import print_function
 
 from optparse import OptionParser
+import os
+import sys
 
-from click.build import ClickBuilder
+from click.build import ClickBuildError, ClickBuilder
 
 
 def run(argv):
@@ -31,8 +33,18 @@ def run(argv):
     if len(args) < 1:
         parser.error("need directory")
     directory = args[0]
+    if not os.path.isdir(directory):
+        parser.error('directory "%s" does not exist' % directory)
+    if not os.path.exists(os.path.join(directory, options.manifest)):
+        parser.error(
+            'directory "%s" does not contain manifest file "%s"' %
+            (directory, options.manifest))
     builder = ClickBuilder()
     builder.add_file(directory, "./")
-    path = builder.build(".", manifest_path=options.manifest)
+    try:
+        path = builder.build(".", manifest_path=options.manifest)
+    except ClickBuildError as e:
+        print(e, file=sys.stderr)
+        return 1
     print("Successfully built package in '%s'." % path)
     return 0
