@@ -19,17 +19,17 @@
 from __future__ import print_function
 
 from optparse import OptionParser
+import sys
 
-from click.paths import default_root
+from click.database import ClickDB
 from click.query import find_package_directory
 from click.user import ClickUser
 
 
 def run(argv):
-    parser = OptionParser("%prog pkgdir {PACKAGE-NAME|PATH}")
+    parser = OptionParser("%prog pkgdir [options] {PACKAGE-NAME|PATH}")
     parser.add_option(
-        "--root", metavar="PATH", default=default_root,
-        help="set top-level directory to PATH (default: %s)" % default_root)
+        "--root", metavar="PATH", help="look for additional packages in PATH")
     parser.add_option(
         "--user", metavar="USER",
         help="look up PACKAGE-NAME for USER (if you have permission; "
@@ -37,10 +37,15 @@ def run(argv):
     options, args = parser.parse_args(argv)
     if len(args) < 1:
         parser.error("need package name")
-    if "/" in args[0]:
-        print(find_package_directory(args[0]))
-    else:
-        package_name = args[0]
-        registry = ClickUser(options.root, user=options.user)
-        print(registry.path(package_name))
+    try:
+        if "/" in args[0]:
+            print(find_package_directory(args[0]))
+        else:
+            db = ClickDB(options.root)
+            package_name = args[0]
+            registry = ClickUser(db, user=options.user)
+            print(registry.path(package_name))
+    except Exception as e:
+        print(e, file=sys.stderr)
+        return 1
     return 0
