@@ -20,6 +20,9 @@ from __future__ import print_function
 __metaclass__ = type
 __all__ = [
     'ClickInstaller',
+    'ClickInstallerAuditError',
+    'ClickInstallerError',
+    'ClickInstallerPermissionDenied',
     ]
 
 
@@ -29,8 +32,11 @@ import inspect
 import json
 import os
 import pwd
+import shutil
+import stat
 import subprocess
 import sys
+import tempfile
 
 from contextlib import closing
 
@@ -61,13 +67,21 @@ except AttributeError:
             self.data._DebPart__member.close()
 
 
-class ClickInstallerPermissionDenied(Exception):
+class ClickInstallerError(Exception):
+    pass
+
+
+class ClickInstallerPermissionDenied(ClickInstallerError):
+    pass
+
+
+class ClickInstallerAuditError(ClickInstallerError):
     pass
 
 
 class ClickInstaller:
-    def __init__(self, root, force_missing_framework=False):
-        self.root = root
+    def __init__(self, db, force_missing_framework=False):
+        self.db = db
         self.force_missing_framework = force_missing_framework
 
     def _preload_path(self):
@@ -85,86 +99,133 @@ class ClickInstaller:
         return os.path.exists(os.path.join(
             frameworks_dir, "%s.framework" % name))
 
-    def audit_control(self, control_part):
-        """Check that all requirements on the control part are met.
+    def _dpkg_architecture(self):
+        return subprocess.check_output(
+            ["dpkg", "--print-architecture"],
+            universal_newlines=True).rstrip("\n")
 
-        Returns the package name.
-        """
-        control_fields = control_part.debcontrol()
+    def extract(self, path, target):
+        command = ["dpkg-deb", "-R", path, target]
+        with open(path, "rb") as fd:
+            env = dict(os.environ)
+            preloads = [self._preload_path()]
+            if "LD_PRELOAD" in env:
+                preloads.append(env["LD_PRELOAD"])
+            env["LD_PRELOAD"] = " ".join(preloads)
+            env["CLICK_BASE_DIR"] = target
+            env["CLICK_PACKAGE_PATH"] = path
+            env["CLICK_PACKAGE_FD"] = str(fd.fileno())
+            env.pop("HOME", None)
+            kwargs = {}
+            if sys.version >= "3.2":
+                kwargs["pass_fds"] = (fd.fileno(),)
+            subprocess.check_call(command, env=env, **kwargs)
 
-        try:
-            click_version = Version(control_fields["Click-Version"])
-        except KeyError:
-            raise ValueError("No Click-Version field")
-        if click_version > spec_version:
-            raise ValueError(
-                "Click-Version: %s newer than maximum supported version %s" %
-                (click_version, spec_version))
-
-        for field in (
-            "Pre-Depends", "Depends", "Recommends", "Suggests", "Enhances",
-            "Conflicts", "Breaks",
-            "Provides",
-        ):
-            if field in control_fields:
-                raise ValueError(
-                    "%s field is forbidden in Click packages" % field)
-
-        scripts = control_part.scripts()
-        if ("preinst" in scripts and
-                static_preinst_matches(scripts["preinst"])):
-            scripts.pop("preinst", None)
-        if scripts:
-            raise ValueError(
-                "Maintainer scripts are forbidden in Click packages "
-                "(found: %s)" %
-                " ".join(sorted(scripts)))
-
-        if not control_part.has_file("manifest"):
-            raise ValueError("Package has no manifest")
-        with control_part.get_file("manifest", encoding="UTF-8") as f:
-            manifest = json.loads(f.read())
-
-        try:
-            package_name = manifest["name"]
-        except KeyError:
-            raise ValueError('No "name" entry in manifest')
-        # TODO: perhaps just do full name validation?
-        if "/" in package_name:
-            raise ValueError(
-                'Invalid character "/" in "name" entry: %s' % package_name)
-        if "_" in package_name:
-            raise ValueError(
-                'Invalid character "_" in "name" entry: %s' % package_name)
-
-        try:
-            package_version = manifest["version"]
-        except KeyError:
-            raise ValueError('No "version" entry in manifest')
-        # TODO: perhaps just do full version validation?
-        if "/" in package_version:
-            raise ValueError(
-                'Invalid character "/" in "version" entry: %s' %
-                package_version)
-        if "_" in package_version:
-            raise ValueError(
-                'Invalid character "_" in "version" entry: %s' %
-                package_version)
-
-        try:
-            framework = manifest["framework"]
-        except KeyError:
-            raise ValueError('No "framework" entry in manifest')
-        if (not self.force_missing_framework and
-                not self._has_framework(framework)):
-            raise ValueError(
-                'Framework "%s" not present on system' % framework)
-
-        return package_name, package_version
-
-    def audit(self, path):
+    def audit(self, path, slow=False, check_arch=False):
         with closing(DebFile(filename=path)) as package:
-            return self.audit_control(package.control)
+            control_fields = package.control.debcontrol()
+
+            try:
+                click_version = Version(control_fields["Click-Version"])
+            except KeyError:
+                raise ClickInstallerAuditError("No Click-Version field")
+            if click_version > spec_version:
+                raise ClickInstallerAuditError(
+                    "Click-Version: %s newer than maximum supported version "
+                    "%s" % (click_version, spec_version))
+
+            for field in (
+                "Pre-Depends", "Depends", "Recommends", "Suggests", "Enhances",
+                "Conflicts", "Breaks",
+                "Provides",
+            ):
+                if field in control_fields:
+                    raise ClickInstallerAuditError(
+                        "%s field is forbidden in Click packages" % field)
+
+            scripts = package.control.scripts()
+            if ("preinst" in scripts and
+                    static_preinst_matches(scripts["preinst"])):
+                scripts.pop("preinst", None)
+            if scripts:
+                raise ClickInstallerAuditError(
+                    "Maintainer scripts are forbidden in Click packages "
+                    "(found: %s)" %
+                    " ".join(sorted(scripts)))
+
+            if not package.control.has_file("manifest"):
+                raise ClickInstallerAuditError("Package has no manifest")
+            with package.control.get_file("manifest", encoding="UTF-8") as f:
+                manifest = json.load(f)
+
+            try:
+                package_name = manifest["name"]
+            except KeyError:
+                raise ClickInstallerAuditError('No "name" entry in manifest')
+            # TODO: perhaps just do full name validation?
+            if "/" in package_name:
+                raise ClickInstallerAuditError(
+                    'Invalid character "/" in "name" entry: %s' % package_name)
+            if "_" in package_name:
+                raise ClickInstallerAuditError(
+                    'Invalid character "_" in "name" entry: %s' % package_name)
+
+            try:
+                package_version = manifest["version"]
+            except KeyError:
+                raise ClickInstallerAuditError(
+                    'No "version" entry in manifest')
+            # TODO: perhaps just do full version validation?
+            if "/" in package_version:
+                raise ClickInstallerAuditError(
+                    'Invalid character "/" in "version" entry: %s' %
+                    package_version)
+            if "_" in package_version:
+                raise ClickInstallerAuditError(
+                    'Invalid character "_" in "version" entry: %s' %
+                    package_version)
+
+            try:
+                framework = manifest["framework"]
+            except KeyError:
+                raise ClickInstallerAuditError(
+                    'No "framework" entry in manifest')
+            if (not self.force_missing_framework and
+                    not self._has_framework(framework)):
+                raise ClickInstallerAuditError(
+                    'Framework "%s" not present on system (use '
+                    '--force-missing-framework option to override)' %
+                    framework)
+
+            if check_arch:
+                architecture = manifest.get("architecture", "all")
+                if architecture != "all":
+                    dpkg_architecture = self._dpkg_architecture()
+                    if isinstance(architecture, list):
+                        if dpkg_architecture not in architecture:
+                            raise ClickInstallerAuditError(
+                                'Package architectures "%s" not compatible '
+                                'with system architecture "%s"' %
+                                (" ".join(architecture), dpkg_architecture))
+                    elif architecture != dpkg_architecture:
+                        raise ClickInstallerAuditError(
+                            'Package architecture "%s" not compatible '
+                            'with system architecture "%s"' %
+                            (architecture, dpkg_architecture))
+
+            if slow:
+                temp_dir = tempfile.mkdtemp(prefix="click")
+                try:
+                    self.extract(path, temp_dir)
+                    command = [
+                        "md5sum", "-c", "--quiet",
+                        os.path.join("DEBIAN", "md5sums"),
+                    ]
+                    subprocess.check_call(command, cwd=temp_dir)
+                finally:
+                    shutil.rmtree(temp_dir)
+
+            return package_name, package_version
 
     def _drop_privileges(self, username):
         if os.geteuid() != 0:
@@ -180,6 +241,7 @@ class ClickInstaller:
         os.setresuid(pw.pw_uid, pw.pw_uid, pw.pw_uid)
         assert os.getresuid() == (pw.pw_uid, pw.pw_uid, pw.pw_uid)
         assert os.getresgid() == (pw.pw_gid, pw.pw_gid, pw.pw_gid)
+        os.umask(0o022)
 
     def _euid_access(self, username, path, mode):
         """Like os.access, but for the effective UID."""
@@ -205,7 +267,8 @@ class ClickInstaller:
                 break
         if not self._euid_access("clickpkg", path, os.W_OK):
             raise ClickInstallerPermissionDenied(
-                "No permission to write to %s as clickpkg user" % path)
+                'Cannot acquire permission to write to %s; either run as root '
+                'with --user, or use "pkcon install-local" instead' % path)
 
     def _install_preexec(self, inst_dir):
         self._drop_privileges("clickpkg")
@@ -221,14 +284,14 @@ class ClickInstaller:
             os.mkdir(os.path.join(admin_dir, "updates"))
             os.mkdir(os.path.join(admin_dir, "triggers"))
 
-    def install(self, path, user=None):
-        package_name, package_version = self.audit(path)
-        package_dir = os.path.join(self.root, package_name)
+    def install(self, path, user=None, all_users=False):
+        package_name, package_version = self.audit(path, check_arch=True)
+        package_dir = os.path.join(self.db.overlay, package_name)
         inst_dir = os.path.join(package_dir, package_version)
-        assert os.path.dirname(os.path.dirname(inst_dir)) == self.root
+        assert os.path.dirname(os.path.dirname(inst_dir)) == self.db.overlay
 
-        self._check_write_permissions(self.root)
-        root_click = os.path.join(self.root, ".click")
+        self._check_write_permissions(self.db.overlay)
+        root_click = os.path.join(self.db.overlay, ".click")
         if not os.path.exists(root_click):
             os.makedirs(root_click)
             if os.geteuid() == 0:
@@ -238,7 +301,13 @@ class ClickInstaller:
         # TODO: sandbox so that this can only write to the unpack directory
         command = [
             "dpkg",
+            # We normally run dpkg as non-root.
             "--force-not-root",
+            # /sbin and /usr/sbin may not necessarily be on $PATH; we don't
+            # use the tools dpkg gets from there.
+            "--force-bad-path",
+            # We check the package architecture ourselves in audit().
+            "--force-architecture",
             "--instdir", inst_dir,
             "--admindir", os.path.join(inst_dir, ".click"),
             "--path-exclude", "*/.click/*",
@@ -252,7 +321,7 @@ class ClickInstaller:
             if "LD_PRELOAD" in env:
                 preloads.append(env["LD_PRELOAD"])
             env["LD_PRELOAD"] = " ".join(preloads)
-            env["CLICK_BASE_DIR"] = self.root
+            env["CLICK_BASE_DIR"] = self.db.overlay
             env["CLICK_PACKAGE_PATH"] = path
             env["CLICK_PACKAGE_FD"] = str(fd.fileno())
             env.pop("HOME", None)
@@ -262,6 +331,18 @@ class ClickInstaller:
             subprocess.check_call(
                 command, preexec_fn=partial(self._install_preexec, inst_dir),
                 env=env, **kwargs)
+        for dirpath, dirnames, filenames in os.walk(inst_dir):
+            for entry in dirnames + filenames:
+                entry_path = os.path.join(dirpath, entry)
+                entry_mode = os.stat(entry_path).st_mode
+                new_entry_mode = entry_mode | stat.S_IRGRP | stat.S_IROTH
+                if entry_mode & stat.S_IXUSR:
+                    new_entry_mode |= stat.S_IXGRP | stat.S_IXOTH
+                if new_entry_mode != entry_mode:
+                    try:
+                        os.chmod(entry_path, new_entry_mode)
+                    except OSError:
+                        pass
 
         current_path = os.path.join(package_dir, "current")
 
@@ -272,7 +353,7 @@ class ClickInstaller:
         else:
             old_version = None
         package_install_hooks(
-            self.root, package_name, old_version, package_version)
+            self.db, package_name, old_version, package_version)
 
         new_path = os.path.join(package_dir, "current.new")
         osextras.symlink_force(package_version, new_path)
@@ -284,8 +365,9 @@ class ClickInstaller:
             os.chown(new_path, pw.pw_uid, pw.pw_gid, follow_symlinks=False)
         os.rename(new_path, current_path)
 
-        if user is not None:
-            registry = ClickUser(self.root, user)
-            registry[package_name] = package_version
+        if user is not None or all_users:
+            registry = ClickUser(self.db, user=user, all_users=all_users)
+            registry.set_version(package_name, package_version)
 
-        # TODO: garbage-collect old directories
+        if old_version is not None:
+            self.db.maybe_remove(package_name, old_version)
