@@ -31,7 +31,7 @@ import subprocess
 import tarfile
 from textwrap import dedent
 
-from click.build import ClickBuilder, ClickSourceBuilder
+from click.build import ClickBuildError, ClickBuilder, ClickSourceBuilder
 from click.preinst import static_preinst
 from click.tests.helpers import TestCase, mkfile, touch
 
@@ -94,6 +94,19 @@ class TestClickBuilderBaseMixin:
             self.builder.read_manifest(manifest_path)
             self.assertEqual(epochless_version, self.builder.epochless_version)
 
+    def test_manifest_syntax_error(self):
+        self.use_temp_dir()
+        manifest_path = os.path.join(self.temp_dir, "manifest.json")
+        with mkfile(manifest_path) as manifest:
+            # The comma after the "name" entry is intentionally missing.
+            print(dedent("""\
+                {
+                    "name": "com.ubuntu.test"
+                    "version": "1.0"
+                }"""), file=manifest)
+        self.assertRaises(
+            ClickBuildError, self.builder.read_manifest, manifest_path)
+
 
 class TestClickBuilder(TestCase, TestClickBuilderBaseMixin):
     def setUp(self):
@@ -112,17 +125,18 @@ class TestClickBuilder(TestCase, TestClickBuilderBaseMixin):
         with mkfile(os.path.join(scratch, "bin", "foo")) as f:
             f.write("test /bin/foo\n")
         os.symlink("foo", os.path.join(scratch, "bin", "bar"))
+        touch(os.path.join(scratch, ".git", "config"))
         with mkfile(os.path.join(scratch, "toplevel")) as f:
             f.write("test /toplevel\n")
         with mkfile(os.path.join(scratch, "manifest.json")) as f:
-            f.write(json.dumps({
+            json.dump({
                 "name": "com.ubuntu.test",
                 "version": "1.0",
                 "maintainer": "Foo Bar <foo@example.org>",
                 "title": "test title",
                 "architecture": "all",
                 "framework": "ubuntu-sdk-13.10",
-            }))
+            }, f)
             # build() overrides this back to 0o644
             os.fchmod(f.fileno(), 0o600)
         self.builder.add_file(scratch, "/")
@@ -132,7 +146,7 @@ class TestClickBuilder(TestCase, TestClickBuilderBaseMixin):
         for key, value in (
             ("Package", "com.ubuntu.test"),
             ("Version", "1.0"),
-            ("Click-Version", "0.2"),
+            ("Click-Version", "0.4"),
             ("Architecture", "all"),
             ("Maintainer", "Foo Bar <foo@example.org>"),
             ("Description", "test title"),
@@ -146,7 +160,11 @@ class TestClickBuilder(TestCase, TestClickBuilderBaseMixin):
         self.assertEqual(0o644, stat.S_IMODE(os.stat(manifest_path).st_mode))
         with open(os.path.join(scratch, "manifest.json")) as source, \
                 open(manifest_path) as target:
-            self.assertEqual(source.read(), target.read())
+            source_json = json.load(source)
+            target_json = json.load(target)
+            self.assertNotEqual("", target_json["installed-size"])
+            del target_json["installed-size"]
+            self.assertEqual(source_json, target_json)
         with open(os.path.join(control_path, "md5sums")) as md5sums:
             self.assertRegex(
                 md5sums.read(),
@@ -186,20 +204,47 @@ class TestClickBuilder(TestCase, TestClickBuilderBaseMixin):
         scratch = os.path.join(self.temp_dir, "scratch")
         touch(os.path.join(scratch, ".click", "evil-file"))
         with mkfile(os.path.join(scratch, "manifest.json")) as f:
-            f.write(json.dumps({
+            json.dump({
                 "name": "com.ubuntu.test",
                 "version": "1.0",
                 "maintainer": "Foo Bar <foo@example.org>",
                 "title": "test title",
                 "architecture": "all",
                 "framework": "ubuntu-sdk-13.10",
-            }))
+            }, f)
         self.builder.add_file(scratch, "/")
         path = self.builder.build(self.temp_dir)
         extract_path = os.path.join(self.temp_dir, "extract")
         subprocess.check_call(["dpkg-deb", "-x", path, extract_path])
         self.assertEqual([], os.listdir(extract_path))
 
+    def test_build_multiple_architectures(self):
+        self.use_temp_dir()
+        scratch = os.path.join(self.temp_dir, "scratch")
+        with mkfile(os.path.join(scratch, "manifest.json")) as f:
+            json.dump({
+                "name": "com.ubuntu.test",
+                "version": "1.0",
+                "maintainer": "Foo Bar <foo@example.org>",
+                "title": "test title",
+                "architecture": ["armhf", "i386"],
+                "framework": "ubuntu-sdk-13.10",
+            }, f)
+        self.builder.add_file(scratch, "/")
+        path = os.path.join(self.temp_dir, "com.ubuntu.test_1.0_multi.click")
+        self.assertEqual(path, self.builder.build(self.temp_dir))
+        self.assertTrue(os.path.exists(path))
+        self.assertEqual("multi", self.extract_field(path, "Architecture"))
+        control_path = os.path.join(self.temp_dir, "control")
+        subprocess.check_call(["dpkg-deb", "-e", path, control_path])
+        manifest_path = os.path.join(control_path, "manifest")
+        with open(os.path.join(scratch, "manifest.json")) as source, \
+                open(manifest_path) as target:
+            source_json = json.load(source)
+            target_json = json.load(target)
+            del target_json["installed-size"]
+            self.assertEqual(source_json, target_json)
+
 
 class TestClickSourceBuilder(TestCase, TestClickBuilderBaseMixin):
     def setUp(self):
@@ -213,14 +258,14 @@ class TestClickSourceBuilder(TestCase, TestClickBuilderBaseMixin):
         touch(os.path.join(scratch, "bin", "foo"))
         touch(os.path.join(scratch, ".git", "config"))
         with mkfile(os.path.join(scratch, "manifest.json")) as f:
-            f.write(json.dumps({
+            json.dump({
                 "name": "com.ubuntu.test",
                 "version": "1.0",
                 "maintainer": "Foo Bar <foo@example.org>",
                 "title": "test title",
                 "architecture": "all",
                 "framework": "ubuntu-sdk-13.10",
-            }))
+            }, f)
             # build() overrides this back to 0o644
             os.fchmod(f.fileno(), 0o600)
         self.builder.add_file(scratch, "./")
