@@ -24,21 +24,25 @@ __all__ = [
 
 
 from contextlib import contextmanager
+import hashlib
 import json
 import os
+import shutil
 import stat
 import subprocess
 
-from contextlib import closing
 from unittest import skipUnless
 
 from debian.deb822 import Deb822
-# BAW 2013-04-16: Get the DebFile class from here because of compatibility
-# issues.  See the comments in that module for details.
-from click.install import DebFile
 
 from click import install, osextras
-from click.install import ClickInstaller, ClickInstallerPermissionDenied
+from click.build import ClickBuilder
+from click.database import ClickDB
+from click.install import (
+    ClickInstaller,
+    ClickInstallerAuditError,
+    ClickInstallerPermissionDenied,
+)
 from click.preinst import static_preinst
 from click.tests.helpers import TestCase, mkfile, mock, touch
 
@@ -64,38 +68,39 @@ class TestClickInstaller(TestCase):
     def setUp(self):
         super(TestClickInstaller, self).setUp()
         self.use_temp_dir()
+        self.db = ClickDB(self.temp_dir)
 
     def make_fake_package(self, control_fields=None, manifest=None,
                           control_scripts=None, data_files=None):
-        """Build a fake package with given contents.
-
-        We can afford to use dpkg-deb here since it's easy, just for testing.
-        """
+        """Build a fake package with given contents."""
         control_fields = {} if control_fields is None else control_fields
         control_scripts = {} if control_scripts is None else control_scripts
-        data_files = [] if data_files is None else data_files
+        data_files = {} if data_files is None else data_files
 
-        package_dir = os.path.join(self.temp_dir, "fake-package")
-        control_dir = os.path.join(package_dir, "DEBIAN")
+        data_dir = os.path.join(self.temp_dir, "fake-package")
+        control_dir = os.path.join(self.temp_dir, "DEBIAN")
         with mkfile(os.path.join(control_dir, "control")) as control:
             for key, value in control_fields.items():
                 print('%s: %s' % (key.title(), value), file=control)
             print(file=control)
         if manifest is not None:
             with mkfile(os.path.join(control_dir, "manifest")) as f:
-                print(json.dumps(manifest), file=f)
+                json.dump(manifest, f)
+                print(file=f)
         for name, contents in control_scripts.items():
             with mkfile(os.path.join(control_dir, name)) as script:
                 script.write(contents)
-        for name in data_files:
-            touch(os.path.join(package_dir, name))
-        package_path = '%s.click' % package_dir
-        with open("/dev/null", "w") as devnull:
-            env = dict(os.environ)
-            env["NO_PKG_MANGLE"] = "1"
-            subprocess.check_call(
-                ["dpkg-deb", "--nocheck", "-b", package_dir, package_path],
-                stdout=devnull, stderr=devnull, env=env)
+        osextras.ensuredir(data_dir)
+        for name, path in data_files.items():
+            if path is None:
+                touch(os.path.join(data_dir, name))
+            elif os.path.isdir(path):
+                shutil.copytree(path, os.path.join(data_dir, name))
+            else:
+                shutil.copy2(path, os.path.join(data_dir, name))
+        package_path = '%s.click' % data_dir
+        ClickBuilder()._pack(
+            self.temp_dir, control_dir, data_dir, package_path)
         return package_path
 
     @contextmanager
@@ -111,114 +116,100 @@ class TestClickInstaller(TestCase):
                 os.path.join(install.frameworks_dir, "%s.framework" % name))
             install.frameworks_dir = old_dir
 
-    def test_audit_control_no_click_version(self):
+    def test_audit_no_click_version(self):
         path = self.make_fake_package()
-        with closing(DebFile(filename=path)) as package:
-            self.assertRaisesRegex(
-                ValueError, "No Click-Version field",
-                ClickInstaller(self.temp_dir).audit_control, package.control)
+        self.assertRaisesRegex(
+            ClickInstallerAuditError, "No Click-Version field",
+            ClickInstaller(self.db).audit, path)
 
-    def test_audit_control_bad_click_version(self):
+    def test_audit_bad_click_version(self):
         path = self.make_fake_package(control_fields={"Click-Version": "|"})
-        with closing(DebFile(filename=path)) as package:
-            self.assertRaises(
-                ValueError,
-                ClickInstaller(self.temp_dir).audit_control, package.control)
+        self.assertRaises(ValueError, ClickInstaller(self.db).audit, path)
 
-    def test_audit_control_new_click_version(self):
+    def test_audit_new_click_version(self):
         path = self.make_fake_package(control_fields={"Click-Version": "999"})
-        with closing(DebFile(filename=path)) as package:
-            self.assertRaisesRegex(
-                ValueError,
-                "Click-Version: 999 newer than maximum supported version .*",
-                ClickInstaller(self.temp_dir).audit_control, package.control)
+        self.assertRaisesRegex(
+            ClickInstallerAuditError,
+            "Click-Version: 999 newer than maximum supported version .*",
+            ClickInstaller(self.db).audit, path)
 
-    def test_audit_control_forbids_depends(self):
+    def test_audit_forbids_depends(self):
         path = self.make_fake_package(
             control_fields={
                 "Click-Version": "0.2",
                 "Depends": "libc6",
             })
-        with closing(DebFile(filename=path)) as package, \
-             self.make_framework("ubuntu-sdk-13.10"):
+        with self.make_framework("ubuntu-sdk-13.10"):
             self.assertRaisesRegex(
-                ValueError, "Depends field is forbidden in Click packages",
-                ClickInstaller(self.temp_dir).audit_control, package.control)
+                ClickInstallerAuditError,
+                "Depends field is forbidden in Click packages",
+                ClickInstaller(self.db).audit, path)
 
-    def test_audit_control_forbids_maintscript(self):
+    def test_audit_forbids_maintscript(self):
         path = self.make_fake_package(
             control_fields={"Click-Version": "0.2"},
             control_scripts={
                 "preinst": "#! /bin/sh\n",
                 "postinst": "#! /bin/sh\n",
             })
-        with closing(DebFile(filename=path)) as package, \
-             self.make_framework("ubuntu-sdk-13.10"):
+        with self.make_framework("ubuntu-sdk-13.10"):
             self.assertRaisesRegex(
-                ValueError,
+                ClickInstallerAuditError,
                 r"Maintainer scripts are forbidden in Click packages "
                 r"\(found: postinst preinst\)",
-                ClickInstaller(self.temp_dir).audit_control, package.control)
+                ClickInstaller(self.db).audit, path)
 
-    def test_audit_control_requires_manifest(self):
+    def test_audit_requires_manifest(self):
         path = self.make_fake_package(
             control_fields={"Click-Version": "0.2"},
             control_scripts={"preinst": static_preinst})
-        with closing(DebFile(filename=path)) as package, \
-             self.make_framework("ubuntu-sdk-13.10"):
+        with self.make_framework("ubuntu-sdk-13.10"):
             self.assertRaisesRegex(
-                ValueError, "Package has no manifest",
-                ClickInstaller(self.temp_dir).audit_control, package.control)
+                ClickInstallerAuditError, "Package has no manifest",
+                ClickInstaller(self.db).audit, path)
 
-    def test_audit_control_invalid_manifest_json(self):
+    def test_audit_invalid_manifest_json(self):
         path = self.make_fake_package(
             control_fields={"Click-Version": "0.2"},
             control_scripts={"manifest": "{", "preinst": static_preinst})
-        with closing(DebFile(filename=path)) as package, \
-             self.make_framework("ubuntu-sdk-13.10"):
-            self.assertRaises(
-                ValueError,
-                ClickInstaller(self.temp_dir).audit_control, package.control)
+        with self.make_framework("ubuntu-sdk-13.10"):
+            self.assertRaises(ValueError, ClickInstaller(self.db).audit, path)
 
-    def test_audit_control_no_name(self):
+    def test_audit_no_name(self):
         path = self.make_fake_package(
             control_fields={"Click-Version": "0.2"},
             manifest={})
-        with closing(DebFile(filename=path)) as package:
-            self.assertRaisesRegex(
-                ValueError, 'No "name" entry in manifest',
-                ClickInstaller(self.temp_dir).audit_control, package.control)
+        self.assertRaisesRegex(
+            ClickInstallerAuditError, 'No "name" entry in manifest',
+            ClickInstaller(self.db).audit, path)
 
-    def test_audit_control_name_bad_character(self):
+    def test_audit_name_bad_character(self):
         path = self.make_fake_package(
             control_fields={"Click-Version": "0.2"},
             manifest={"name": "../evil"})
-        with closing(DebFile(filename=path)) as package:
-            self.assertRaisesRegex(
-                ValueError,
-                'Invalid character "/" in "name" entry: ../evil',
-                ClickInstaller(self.temp_dir).audit_control, package.control)
+        self.assertRaisesRegex(
+            ClickInstallerAuditError,
+            'Invalid character "/" in "name" entry: ../evil',
+            ClickInstaller(self.db).audit, path)
 
-    def test_audit_control_no_version(self):
+    def test_audit_no_version(self):
         path = self.make_fake_package(
             control_fields={"Click-Version": "0.2"},
             manifest={"name": "test-package"})
-        with closing(DebFile(filename=path)) as package:
-            self.assertRaisesRegex(
-                ValueError, 'No "version" entry in manifest',
-                ClickInstaller(self.temp_dir).audit_control, package.control)
+        self.assertRaisesRegex(
+            ClickInstallerAuditError, 'No "version" entry in manifest',
+            ClickInstaller(self.db).audit, path)
 
-    def test_audit_control_no_framework(self):
+    def test_audit_no_framework(self):
         path = self.make_fake_package(
             control_fields={"Click-Version": "0.2"},
             manifest={"name": "test-package", "version": "1.0"},
             control_scripts={"preinst": static_preinst})
-        with closing(DebFile(filename=path)) as package:
-            self.assertRaisesRegex(
-                ValueError, 'No "framework" entry in manifest',
-                ClickInstaller(self.temp_dir).audit_control, package.control)
+        self.assertRaisesRegex(
+            ClickInstallerAuditError, 'No "framework" entry in manifest',
+            ClickInstaller(self.db).audit, path)
 
-    def test_audit_control_missing_framework(self):
+    def test_audit_missing_framework(self):
         path = self.make_fake_package(
             control_fields={"Click-Version": "0.2"},
             manifest={
@@ -227,13 +218,13 @@ class TestClickInstaller(TestCase):
                 "framework": "missing",
             },
             control_scripts={"preinst": static_preinst})
-        with closing(DebFile(filename=path)) as package, \
-             self.make_framework("present"):
+        with self.make_framework("present"):
             self.assertRaisesRegex(
-                ValueError, 'Framework "missing" not present on system',
-                ClickInstaller(self.temp_dir).audit_control, package.control)
+                ClickInstallerAuditError,
+                'Framework "missing" not present on system.*',
+                ClickInstaller(self.db).audit, path)
 
-    def test_audit_control_missing_framework_force(self):
+    def test_audit_missing_framework_force(self):
         path = self.make_fake_package(
             control_fields={"Click-Version": "0.2"},
             manifest={
@@ -241,9 +232,8 @@ class TestClickInstaller(TestCase):
                 "version": "1.0",
                 "framework": "missing",
             })
-        with closing(DebFile(filename=path)) as package, \
-             self.make_framework("present"):
-            ClickInstaller(self.temp_dir, True).audit_control(package.control)
+        with self.make_framework("present"):
+            ClickInstaller(self.db, True).audit(path)
 
     def test_audit_passes_correct_package(self):
         path = self.make_fake_package(
@@ -255,9 +245,53 @@ class TestClickInstaller(TestCase):
             },
             control_scripts={"preinst": static_preinst})
         with self.make_framework("ubuntu-sdk-13.10"):
-            installer = ClickInstaller(self.temp_dir)
+            installer = ClickInstaller(self.db)
             self.assertEqual(("test-package", "1.0"), installer.audit(path))
 
+    def test_audit_broken_md5sums(self):
+        path = self.make_fake_package(
+            control_fields={"Click-Version": "0.2"},
+            manifest={
+                "name": "test-package",
+                "version": "1.0",
+                "framework": "ubuntu-sdk-13.10",
+            },
+            control_scripts={
+                "preinst": static_preinst,
+                "md5sums": "%s  foo" % ("0" * 32),
+            },
+            data_files={"foo": None})
+        with self.make_framework("ubuntu-sdk-13.10"), \
+             mock_quiet_subprocess_call():
+            installer = ClickInstaller(self.db)
+            self.assertRaises(
+                subprocess.CalledProcessError, installer.audit,
+                path, slow=True)
+
+    def test_audit_matching_md5sums(self):
+        data_path = os.path.join(self.temp_dir, "foo")
+        with mkfile(data_path) as data:
+            print("test", file=data)
+        with open(data_path, "rb") as data:
+            data_md5sum = hashlib.md5(data.read()).hexdigest()
+        path = self.make_fake_package(
+            control_fields={"Click-Version": "0.2"},
+            manifest={
+                "name": "test-package",
+                "version": "1.0",
+                "framework": "ubuntu-sdk-13.10",
+            },
+            control_scripts={
+                "preinst": static_preinst,
+                "md5sums": "%s  foo" % data_md5sum,
+            },
+            data_files={"foo": data_path})
+        with self.make_framework("ubuntu-sdk-13.10"), \
+             mock_quiet_subprocess_call():
+            installer = ClickInstaller(self.db)
+            self.assertEqual(
+                ("test-package", "1.0"), installer.audit(path, slow=True))
+
     def test_no_write_permission(self):
         path = self.make_fake_package(
             control_fields={"Click-Version": "0.2"},
@@ -269,7 +303,7 @@ class TestClickInstaller(TestCase):
             control_scripts={"preinst": static_preinst})
         write_mask = ~(stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH)
         with self.make_framework("ubuntu-sdk-13.10"):
-            installer = ClickInstaller(self.temp_dir)
+            installer = ClickInstaller(self.db)
             temp_dir_mode = os.stat(self.temp_dir).st_mode
             try:
                 os.chmod(self.temp_dir, temp_dir_mode & write_mask)
@@ -298,9 +332,10 @@ class TestClickInstaller(TestCase):
                 "framework": "ubuntu-sdk-13.10",
             },
             control_scripts={"preinst": static_preinst},
-            data_files=["foo"])
+            data_files={"foo": None})
         root = os.path.join(self.temp_dir, "root")
-        installer = ClickInstaller(root)
+        db = ClickDB(root)
+        installer = ClickInstaller(db)
         with self.make_framework("ubuntu-sdk-13.10"), \
              mock_quiet_subprocess_call():
             installer.install(path)
@@ -313,7 +348,9 @@ class TestClickInstaller(TestCase):
         self.assertCountEqual([".click", "foo"], os.listdir(inst_dir))
         status_path = os.path.join(inst_dir, ".click", "status")
         with open(status_path) as status_file:
-            status = list(Deb822.iter_paragraphs(status_file))
+            # .readlines() avoids the need for a python-apt backport to
+            # Ubuntu 12.04 LTS.
+            status = list(Deb822.iter_paragraphs(status_file.readlines()))
         self.assertEqual(1, len(status))
         self.assertEqual({
             "Package": "test-package",
@@ -325,7 +362,7 @@ class TestClickInstaller(TestCase):
             "Click-Version": "0.2",
         }, status[0])
         mock_package_install_hooks.assert_called_once_with(
-            root, "test-package", None, "1.0")
+            db, "test-package", None, "1.0")
 
     @skipUnless(
         os.path.exists(ClickInstaller(None)._preload_path()),
@@ -359,9 +396,9 @@ class TestClickInstaller(TestCase):
                 "framework": "ubuntu-sdk-13.10",
             },
             control_scripts={"preinst": static_preinst},
-            data_files=["foo"])
+            data_files={"foo": None})
         root = os.path.join(self.temp_dir, "root")
-        installer = ClickInstaller(root)
+        installer = ClickInstaller(ClickDB(root))
         with self.make_framework("ubuntu-sdk-13.10"), \
              mock.patch("subprocess.call") as mock_call:
             mock_call.side_effect = call_side_effect
@@ -390,25 +427,27 @@ class TestClickInstaller(TestCase):
                 "framework": "ubuntu-sdk-13.10",
             },
             control_scripts={"preinst": static_preinst},
-            data_files=["foo"])
+            data_files={"foo": None})
         root = os.path.join(self.temp_dir, "root")
         package_dir = os.path.join(root, "test-package")
         inst_dir = os.path.join(package_dir, "current")
         os.makedirs(os.path.join(package_dir, "1.0"))
         os.symlink("1.0", inst_dir)
-        installer = ClickInstaller(root)
+        db = ClickDB(root)
+        installer = ClickInstaller(db)
         with self.make_framework("ubuntu-sdk-13.10"), \
              mock_quiet_subprocess_call():
             installer.install(path)
         self.assertCountEqual([".click", "test-package"], os.listdir(root))
-        self.assertCountEqual(
-            ["1.0", "1.1", "current"], os.listdir(package_dir))
+        self.assertCountEqual(["1.1", "current"], os.listdir(package_dir))
         self.assertTrue(os.path.islink(inst_dir))
         self.assertEqual("1.1", os.readlink(inst_dir))
         self.assertCountEqual([".click", "foo"], os.listdir(inst_dir))
         status_path = os.path.join(inst_dir, ".click", "status")
         with open(status_path) as status_file:
-            status = list(Deb822.iter_paragraphs(status_file))
+            # .readlines() avoids the need for a python-apt backport to
+            # Ubuntu 12.04 LTS.
+            status = list(Deb822.iter_paragraphs(status_file.readlines()))
         self.assertEqual(1, len(status))
         self.assertEqual({
             "Package": "test-package",
@@ -420,4 +459,117 @@ class TestClickInstaller(TestCase):
             "Click-Version": "0.2",
         }, status[0])
         mock_package_install_hooks.assert_called_once_with(
-            root, "test-package", "1.0", "1.1")
+            db, "test-package", "1.0", "1.1")
+
+    def _get_mode(self, path):
+        return stat.S_IMODE(os.stat(path).st_mode)
+
+    @skipUnless(
+        os.path.exists(ClickInstaller(None)._preload_path()),
+        "preload bits not built; installing packages will fail")
+    @mock.patch("click.install.package_install_hooks")
+    def test_world_readable(self, mock_package_install_hooks):
+        owner_only_file = os.path.join(self.temp_dir, "owner-only-file")
+        touch(owner_only_file)
+        os.chmod(owner_only_file, stat.S_IRUSR | stat.S_IWUSR)
+        owner_only_dir = os.path.join(self.temp_dir, "owner-only-dir")
+        os.mkdir(owner_only_dir, stat.S_IRWXU)
+        path = self.make_fake_package(
+            control_fields={
+                "Package": "test-package",
+                "Version": "1.1",
+                "Architecture": "all",
+                "Maintainer": "Foo Bar <foo@example.org>",
+                "Description": "test",
+                "Click-Version": "0.2",
+            },
+            manifest={
+                "name": "test-package",
+                "version": "1.1",
+                "framework": "ubuntu-sdk-13.10",
+            },
+            control_scripts={"preinst": static_preinst},
+            data_files={
+                "world-readable-file": owner_only_file,
+                "world-readable-dir": owner_only_dir,
+            })
+        root = os.path.join(self.temp_dir, "root")
+        db = ClickDB(root)
+        installer = ClickInstaller(db)
+        with self.make_framework("ubuntu-sdk-13.10"), \
+             mock_quiet_subprocess_call():
+            installer.install(path)
+        inst_dir = os.path.join(root, "test-package", "current")
+        self.assertEqual(
+            stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH,
+            self._get_mode(os.path.join(inst_dir, "world-readable-file")))
+        self.assertEqual(
+            stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP |
+            stat.S_IROTH | stat.S_IXOTH,
+            self._get_mode(os.path.join(inst_dir, "world-readable-dir")))
+
+    @skipUnless(
+        os.path.exists(ClickInstaller(None)._preload_path()),
+        "preload bits not built; installing packages will fail")
+    @mock.patch("click.install.package_install_hooks")
+    @mock.patch("click.install.ClickInstaller._dpkg_architecture")
+    def test_single_architecture(self, mock_dpkg_architecture,
+                                 mock_package_install_hooks):
+        mock_dpkg_architecture.return_value = "armhf"
+        path = self.make_fake_package(
+            control_fields={
+                "Package": "test-package",
+                "Version": "1.1",
+                "Architecture": "armhf",
+                "Maintainer": "Foo Bar <foo@example.org>",
+                "Description": "test",
+                "Click-Version": "0.2",
+            },
+            manifest={
+                "name": "test-package",
+                "version": "1.1",
+                "framework": "ubuntu-sdk-13.10",
+                "architecture": "armhf",
+            },
+            control_scripts={"preinst": static_preinst})
+        root = os.path.join(self.temp_dir, "root")
+        db = ClickDB(root)
+        installer = ClickInstaller(db)
+        with self.make_framework("ubuntu-sdk-13.10"), \
+             mock_quiet_subprocess_call():
+            installer.install(path)
+        self.assertTrue(
+            os.path.exists(os.path.join(root, "test-package", "current")))
+
+    @skipUnless(
+        os.path.exists(ClickInstaller(None)._preload_path()),
+        "preload bits not built; installing packages will fail")
+    @mock.patch("click.install.package_install_hooks")
+    @mock.patch("click.install.ClickInstaller._dpkg_architecture")
+    def test_multiple_architectures(self, mock_dpkg_architecture,
+                                    mock_package_install_hooks):
+        mock_dpkg_architecture.return_value = "armhf"
+        path = self.make_fake_package(
+            control_fields={
+                "Package": "test-package",
+                "Version": "1.1",
+                "Architecture": "multi",
+                "Maintainer": "Foo Bar <foo@example.org>",
+                "Description": "test",
+                "Click-Version": "0.2",
+            },
+            manifest={
+                "name": "test-package",
+                "version": "1.1",
+                "framework": "ubuntu-sdk-13.10",
+                "architecture": ["armhf", "i386"],
+            },
+            control_scripts={"preinst": static_preinst})
+        root = os.path.join(self.temp_dir, "root")
+        db = ClickDB(root)
+        installer = ClickInstaller(db)
+        with self.make_framework("ubuntu-sdk-13.10"), \
+             mock_quiet_subprocess_call():
+            installer.install(path)
+        self.assertTrue(
+            os.path.exists(os.path.join(root, "test-package", "current")))
